Package: cloudosR
Type: Package
Title: 'Lifebit' Platform 'API' Client
Version: 0.2.0
Authors@R: person("Leila Mansouri", email = "leila.mansouri@lifebit.ai", role = c("aut", "cre"))
Description: Interacts with the 'Lifebit' Platform Cohort Browser 'API'
    <https://cloudos.lifebit.ai>. Enables schema discovery, table exploration,
    and read-only 'SQL' query execution with policy-aware behavior and team-based
    access control for cohort data analysis. Requires bastion-enabled workspaces
    for 'API' access.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 4.1.0)
Imports: httr2, jsonlite, utils
Suggests: testthat (>= 3.0.0), knitr, rmarkdown
VignetteBuilder: knitr
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-28 13:42:42 UTC; leilamansouri
Author: Leila Mansouri [aut, cre]
Maintainer: Leila Mansouri <leila.mansouri@lifebit.ai>
Repository: CRAN
Date/Publication: 2026-06-01 09:10:14 UTC
Built: R 4.6.0; ; 2026-06-01 10:23:34 UTC; unix

Package: cloudosR
Type: Package
Title: 'Lifebit' Platform 'API' Client
Version: 0.2.0
Authors@R: person("Leila Mansouri", email = "leila.mansouri@lifebit.ai", role = c("aut", "cre"))
Description: Interacts with the 'Lifebit' Platform Cohort Browser 'API'
    <https://cloudos.lifebit.ai>. Enables schema discovery, table exploration,
    and read-only 'SQL' query execution with policy-aware behavior and team-based
    access control for cohort data analysis. Requires bastion-enabled workspaces
    for 'API' access.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 4.1.0)
Imports: httr2, jsonlite, utils
Suggests: testthat (>= 3.0.0), knitr, rmarkdown
VignetteBuilder: knitr
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-28 13:42:42 UTC; leilamansouri
Author: Leila Mansouri [aut, cre]
Maintainer: Leila Mansouri <leila.mansouri@lifebit.ai>
Repository: CRAN
Date/Publication: 2026-06-01 09:10:14 UTC
Built: R 4.6.0; ; 2026-06-01 10:23:36 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,cloudos_tables)
export(cloudos.cohort_tables)
export(cloudos.configure)
export(cloudos.profile_list)
export(cloudos.query)
export(cloudos.query_results)
export(cloudos.query_status)
export(cloudos.query_submit_async)
export(cloudos.sql_validate)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE---------------------------------------------------------------
# # Install from CRAN (recommended)
# install.packages("cloudosR")
# 
# # Or install development version from GitHub
# devtools::install_github("lifebit-ai/cloudosR")
# 
# # Or install from local source
# install.packages("path/to/cloudosR", repos = NULL, type = "source")

## ----eval=FALSE---------------------------------------------------------------
# library(cloudosR)
# 
# # Configure a profile with your credentials
# cloudos.configure(
#   profilename = "production",
#   apikey = "your-api-key-here",
#   workspace_id = "your-workspace-id",
#   base_url = "https://cloudos.lifebit.ai",  # Optional, this is the default
#   set_default = TRUE  # Set as default profile
# )

## ----eval=FALSE---------------------------------------------------------------
# # With default profile set, you can omit profilename
# results <- cloudos.query(
#   cohort_id = "1a2b3c4d5e6f7g8h9i10j11k",
#   sql = "SELECT * FROM table LIMIT 10"
# )

## ----eval=FALSE---------------------------------------------------------------
# # Production profile
# cloudos.configure(
#   profilename = "production",
#   apikey = "prod-api-key",
#   workspace_id = "prod-workspace-id",
#   set_default = TRUE
# )
# 
# # Staging profile
# cloudos.configure(
#   profilename = "staging",
#   apikey = "staging-api-key",
#   workspace_id = "staging-workspace-id"
# )
# 
# # List all configured profiles
# profiles <- cloudos.profile_list()
# print(profiles)

## ----eval=FALSE---------------------------------------------------------------
# # Simple query (uses default profile)
# results <- cloudos.query(
#   cohort_id = "1a2b3c4d5e6f7g8h9i10j11k",
#   sql = "SELECT person_id, gender_concept_id, birth_datetime
#          FROM person
#          LIMIT 10"
# )
# 
# # View results
# head(results)
# str(results)
# 
# # Access metadata
# attr(results, "total_rows")
# attr(results, "total_pages")

## ----eval=FALSE---------------------------------------------------------------
# # Step 1: Submit query with pagination
# task <- cloudos.query_submit_async(
#   profilename = "production",
#   cohort_id = "1a2b3c4d5e6f7g8h9i10j11k",
#   sql = "SELECT * FROM table LIMIT 10",
#   pagination = list(pageNumber = 0, pageSize = 100)  # Optional
# )
# 
# task_id <- task$task_id
# print(task_id)
# 
# # Step 2: Check status
# status <- cloudos.query_status(
#   profilename = "production",
#   task_id = task_id
# )
# 
# print(status$status)  # "pending", "running", "completed", or "failed"
# print(status$count_of_results)
# 
# # Step 3: Fetch results when completed
# results <- cloudos.query_results(
#   profilename = "production",
#   task_id = task_id
# )

## ----eval=FALSE---------------------------------------------------------------
# # Fetch all pages automatically (default)
# # Note: This submits separate tasks for each page
# all_results <- cloudos.query(
#   cohort_id = "1a2b3c4d5e6f7g8h9i10j11k",
#   sql = "SELECT * FROM person",
#   page_size = 1000,    # Rows per page (default: 1000)
#   all_pages = TRUE     # Fetch all pages (default: TRUE)
# )
# 
# # This will:
# # 1. Submit query for page 0
# # 2. Wait for completion and fetch results
# # 3. Calculate total pages from response metadata
# # 4. Submit separate tasks for pages 1, 2, 3...
# # 5. Wait for all tasks to complete
# # 6. Combine and return all results
# 
# # Fetch only first page
# first_page <- cloudos.query(
#   cohort_id = "1a2b3c4d5e6f7g8h9i10j11k",
#   sql = "SELECT * FROM person",
#   page_size = 100,
#   all_pages = FALSE    # Only fetch first page
# )

## ----eval=FALSE---------------------------------------------------------------
# results <- cloudos.query(
#   cohort_id = "1a2b3c4d5e6f7g8h9i10j11k",
#   sql = "SELECT COUNT(*) FROM person",
#   poll_interval = 5,    # Check every 5 seconds (default: 2)
#   max_wait = 900        # Wait up to 15 minutes (default: 600)
# )

## ----eval=FALSE---------------------------------------------------------------
# # Join multiple tables
# sql_query <- "
#   SELECT
#     p.person_id,
#     p.gender_concept_id,
#     c.condition_concept_id,
#     c.condition_start_date
#   FROM person p
#   JOIN condition_occurrence c ON p.person_id = c.person_id
#   WHERE p.birth_datetime > '1980-01-01'
#   LIMIT 1000
# "
# 
# results <- cloudos.query(
#   cohort_id = "1a2b3c4d5e6f7g8h9i10j11k",
#   sql = sql_query
# )

## ----eval=FALSE---------------------------------------------------------------
# # Authentication errors
# tryCatch({
#   cloudos.query(
#     profilename = "invalid_profile",
#     cohort_id = "123",
#     sql = "SELECT 1"
#   )
# }, error = function(e) {
#   message("Error caught: ", e$message)
# })
# 
# # Query timeout
# tryCatch({
#   cloudos.query(
#     cohort_id = "123",
#     sql = "SELECT * FROM large_table",
#     max_wait = 5  # Very short timeout
#   )
# }, error = function(e) {
#   message("Query timed out: ", e$message)
# })

#' Configure Lifebit Platform Profile
#'
#' Stores API credentials and workspace context for a named profile.
#' This is the required first step before any API wrapper call.
#'
#' @param profilename Character. Name of the profile to create or update.
#' @param apikey Character. API key for authentication.
#' @param workspace_id Character. Workspace/team ID for API requests.
#' @param base_url Character. Base URL for Lifebit Platform API (default: "https://cloudos.lifebit.ai").
#' @param set_default Logical. If TRUE, sets this profile as the default (default: FALSE).
#'
#' @return Invisible NULL. Prints a success message.
#' @export
#'
#' @examples
#' \dontrun{
#'   cloudos.configure(
#'     profilename = "production",
#'     apikey = "your-api-key",
#'     workspace_id = "5c6d3e9bd954e800b23f8c62",
#'     set_default = TRUE
#'   )
#' }
cloudos.configure <- function(profilename = "", 
                              apikey = "", 
                              workspace_id = "",
                              base_url = "https://cloudos.lifebit.ai",
                              set_default = FALSE) {
  
  # Validate required inputs
  if (profilename == "" || is.null(profilename)) {
    stop("Error: profilename is required and cannot be empty.", call. = FALSE)
  }
  
  if (apikey == "" || is.null(apikey)) {
    stop("Error: apikey is required and cannot be empty.", call. = FALSE)
  }
  
  if (workspace_id == "" || is.null(workspace_id)) {
    stop("Error: workspace_id is required and cannot be empty.", call. = FALSE)
  }
  
  # Get config directory path
  config_dir <- get_config_dir()
  
  # Ensure config directory exists
  if (!dir.exists(config_dir)) {
    dir.create(config_dir, recursive = TRUE, mode = "0700")
  }
  
  # Get config file path
  config_file <- get_config_file()
  
  # Read existing config or create new
  if (file.exists(config_file)) {
    config_json <- tryCatch({
      jsonlite::fromJSON(config_file, simplifyVector = FALSE)
    }, error = function(e) {
      list()
    })
  } else {
    config_json <- list()
  }
  
  # If setting as default, unset default flag from other profiles
  if (set_default) {
    for (pname in names(config_json)) {
      if (!is.null(config_json[[pname]]$default)) {
        config_json[[pname]]$default <- FALSE
      }
    }
  }
  
  # Add or update profile
  config_json[[profilename]] <- list(
    apikey = apikey,
    workspace_id = workspace_id,
    base_url = base_url,
    default = set_default,
    created_at = format(Sys.time(), "%Y-%m-%d %H:%M:%S"),
    updated_at = format(Sys.time(), "%Y-%m-%d %H:%M:%S")
  )
  
  # Write config file
  tryCatch({
    jsonlite::write_json(
      config_json, 
      config_file, 
      pretty = TRUE, 
      auto_unbox = TRUE
    )
    
    # Set restrictive permissions on config file (user read/write only)
    Sys.chmod(config_file, mode = "0600")
    
    if (set_default) {
      message(sprintf("Profile '%s' configured successfully and set as default.", profilename))
    } else {
      message(sprintf("Profile '%s' configured successfully.", profilename))
    }
    message(sprintf("Config stored at: %s", config_file))
    
    invisible(NULL)
    
  }, error = function(e) {
    stop(sprintf("Error writing config file: %s", e$message), call. = FALSE)
  })
}


#' List Lifebit Platform Profiles
#'
#' Lists configured profiles so users can confirm available environments.
#'
#' @return Data frame with profile names and metadata (workspace_id, default, created_at, updated_at).
#'         Returns empty data frame if no profiles are configured.
#' @export
#'
#' @examples
#' \dontrun{
#'   profiles <- cloudos.profile_list()
#'   print(profiles)
#' }
cloudos.profile_list <- function() {
  
  config_file <- get_config_file()
  
  # Check if config file exists
  if (!file.exists(config_file)) {
    message("No profiles configured. Use cloudos.configure() to create a profile.")
    return(data.frame(
      profile_name = character(0),
      workspace_id = character(0),
      base_url = character(0),
      default = logical(0),
      created_at = character(0),
      updated_at = character(0),
      stringsAsFactors = FALSE
    ))
  }
  
  # Read config
  config_json <- tryCatch({
    jsonlite::fromJSON(config_file, simplifyVector = FALSE)
  }, error = function(e) {
    stop(sprintf("Error reading config file: %s", e$message), call. = FALSE)
  })
  
  # Check if config is empty
  if (length(config_json) == 0) {
    message("No profiles configured. Use cloudos.configure() to create a profile.")
    return(data.frame(
      profile_name = character(0),
      workspace_id = character(0),
      base_url = character(0),
      default = logical(0),
      created_at = character(0),
      updated_at = character(0),
      stringsAsFactors = FALSE
    ))
  }
  
  # Extract profile metadata
  profiles <- lapply(names(config_json), function(profile_name) {
    profile <- config_json[[profile_name]]
    data.frame(
      profile_name = profile_name,
      workspace_id = profile$workspace_id %||% "",
      base_url = profile$base_url %||% "https://cloudos.lifebit.ai",
      default = profile$default %||% FALSE,
      created_at = profile$created_at %||% "",
      updated_at = profile$updated_at %||% "",
      stringsAsFactors = FALSE
    )
  })
  
  # Combine into single data frame
  profiles_df <- do.call(rbind, profiles)
  
  return(profiles_df)
}


#' Get Config Directory Path
#'
#' Internal function to get the Lifebit Platform config directory path.
#' Uses tools::R_user_dir() for CRAN-compliant persistent storage.
#' Can be overridden with CLOUDOS_CONFIG_DIR environment variable.
#'
#' @return Character. Path to config directory.
#' @keywords internal
get_config_dir <- function() {
  config_dir <- Sys.getenv("CLOUDOS_CONFIG_DIR", "")
  
  if (config_dir == "") {
    # Use R's user config directory (CRAN-compliant)
    config_dir <- tools::R_user_dir("cloudosR", which = "config")
  }
  
  return(config_dir)
}


#' Get Config File Path
#'
#' Internal function to get the Lifebit Platform config file path.
#' Config file is stored in the user's R config directory as .cloudos_config.json
#' using tools::R_user_dir() for CRAN compliance.
#'
#' @return Character. Path to config file.
#' @keywords internal
get_config_file <- function() {
  return(file.path(get_config_dir(), ".cloudos_config.json"))
}


#' Load Profile Configuration
#'
#' Internal function to load a specific profile's configuration.
#' If profilename is empty or NULL, loads the default profile.
#'
#' @param profilename Character. Name of the profile to load. If empty, loads default profile.
#'
#' @return List with profile configuration (apikey, workspace_id, base_url).
#' @keywords internal
load_profile <- function(profilename = "") {
  
  # If no profile specified, try to load default
  load_default <- (profilename == "" || is.null(profilename))
  
  config_file <- get_config_file()
  
  if (!file.exists(config_file)) {
    stop(sprintf(
      "No configuration file found. Use cloudos.configure() to create a profile first.\nExpected location: %s",
      config_file
    ), call. = FALSE)
  }
  
  # Read config
  config_json <- tryCatch({
    jsonlite::fromJSON(config_file, simplifyVector = FALSE)
  }, error = function(e) {
    stop(sprintf("Error reading config file: %s", e$message), call. = FALSE)
  })
  
  # If loading default profile, find it
  if (load_default) {
    default_profile <- NULL
    for (pname in names(config_json)) {
      if (isTRUE(config_json[[pname]]$default)) {
        default_profile <- pname
        profilename <- pname
        break
      }
    }
    
    if (is.null(default_profile)) {
      available_profiles <- paste(names(config_json), collapse = ", ")
      stop(sprintf(
        "No default profile configured.\nAvailable profiles: %s\nSpecify profilename or set a default with cloudos.configure(..., set_default = TRUE)",
        available_profiles
      ), call. = FALSE)
    }
  } else {
    # Check if specified profile exists
    if (!profilename %in% names(config_json)) {
      available_profiles <- paste(names(config_json), collapse = ", ")
      stop(sprintf(
        "Profile '%s' not found.\nAvailable profiles: %s\nUse cloudos.configure() to create this profile.",
        profilename,
        available_profiles
      ), call. = FALSE)
    }
  }
  
  profile <- config_json[[profilename]]
  
  # Validate profile has required fields
  if (is.null(profile$apikey) || profile$apikey == "") {
    stop(sprintf("Profile '%s' is missing apikey.", profilename), call. = FALSE)
  }
  
  if (is.null(profile$workspace_id) || profile$workspace_id == "") {
    stop(sprintf("Profile '%s' is missing workspace_id.", profilename), call. = FALSE)
  }
  
  # Set default base_url if not present
  if (is.null(profile$base_url) || profile$base_url == "") {
    profile$base_url <- "https://cloudos.lifebit.ai"
  }
  
  return(profile)
}

test_that("cloudos.configure creates profile successfully", {
  # Skip if not in an interactive environment or CI
  skip_on_cran()
  skip_if_not(interactive(), "Skipping interactive test")
  
  # Create a temporary config directory for testing
  temp_config <- tempfile("cloudos_test_config")
  dir.create(temp_config, recursive = TRUE)
  
  # Set temp config directory
  old_env <- Sys.getenv("CLOUDOS_CONFIG_DIR", unset = NA)
  Sys.setenv(CLOUDOS_CONFIG_DIR = temp_config)
  
  on.exit({
    # Restore original environment
    if (is.na(old_env)) {
      Sys.unsetenv("CLOUDOS_CONFIG_DIR")
    } else {
      Sys.setenv(CLOUDOS_CONFIG_DIR = old_env)
    }
    # Clean up temp directory
    unlink(temp_config, recursive = TRUE)
  })
  
  # Configure profile
  expect_message(
    cloudos.configure(
      profilename = "test_profile",
      apikey = "test-key-12345",
      workspace_id = "test-workspace-67890",
      set_default = TRUE
    ),
    "configured successfully"
  )
  
  # Check config file exists
  config_file <- file.path(temp_config, ".cloudos_config.json")
  expect_true(file.exists(config_file))
})


test_that("cloudos.profile_list returns empty df when no profiles", {
  # Skip on CRAN
  skip_on_cran()
  
  # Create a temporary config directory for testing
  temp_config <- tempfile("cloudos_test_config")
  dir.create(temp_config, recursive = TRUE)
  
  # Set temp config directory
  old_env <- Sys.getenv("CLOUDOS_CONFIG_DIR", unset = NA)
  Sys.setenv(CLOUDOS_CONFIG_DIR = temp_config)
  
  on.exit({
    # Restore original environment
    if (is.na(old_env)) {
      Sys.unsetenv("CLOUDOS_CONFIG_DIR")
    } else {
      Sys.setenv(CLOUDOS_CONFIG_DIR = old_env)
    }
    # Clean up temp directory
    unlink(temp_config, recursive = TRUE)
  })
  
  expect_message(
    profiles <- cloudos.profile_list(),
    "No profiles configured"
  )
  
  expect_s3_class(profiles, "data.frame")
  expect_equal(nrow(profiles), 0)
})


test_that("load_profile fails with helpful error when no config", {
  # Skip on CRAN
  skip_on_cran()
  
  # Create a temporary config directory for testing
  temp_config <- tempfile("cloudos_test_config")
  dir.create(temp_config, recursive = TRUE)
  
  # Set temp config directory
  old_env <- Sys.getenv("CLOUDOS_CONFIG_DIR", unset = NA)
  Sys.setenv(CLOUDOS_CONFIG_DIR = temp_config)
  
  on.exit({
    # Restore original environment
    if (is.na(old_env)) {
      Sys.unsetenv("CLOUDOS_CONFIG_DIR")
    } else {
      Sys.setenv(CLOUDOS_CONFIG_DIR = old_env)
    }
    # Clean up temp directory
    unlink(temp_config, recursive = TRUE)
  })
  
  expect_error(
    cloudosR:::load_profile("test_profile"),
    "No configuration file found"
  )
})


test_that("validate_required_string catches empty strings", {
  expect_error(
    cloudosR:::validate_required_string("", "test_param"),
    "test_param is required"
  )
  
  expect_error(
    cloudosR:::validate_required_string(NULL, "test_param"),
    "test_param is required"
  )
})

#' Null-coalescing Operator
#'
#' Returns the left-hand side if it's not NULL, otherwise returns the right-hand side.
#'
#' @param x First value to check.
#' @param y Default value if x is NULL.
#'
#' @return x if not NULL, otherwise y.
#' @name null-coalesce
#' @keywords internal
`%||%` <- function(x, y) {
  if (is.null(x)) y else x
}


#' Validate Required String Parameter
#'
#' Internal function to validate that a string parameter is not empty or NULL.
#'
#' @param value Character. Value to validate.
#' @param param_name Character. Name of the parameter (for error messages).
#'
#' @return NULL (throws error if validation fails).
#' @keywords internal
validate_required_string <- function(value, param_name) {
  if (is.null(value) || length(value) == 0 || value == "") {
    stop(sprintf("Error: %s is required and cannot be empty.", param_name), call. = FALSE)
  }
  invisible(NULL)
}


#' Handle API Error Response
#'
#' Internal function to handle API error responses and generate user-friendly error messages.
#'
#' @param response HTTP response object from httr.
#' @param endpoint Character. Endpoint that was called (for error context).
#'
#' @return NULL (throws error with formatted message).
#' @keywords internal
handle_api_error <- function(response, endpoint) {
  status <- httr2::resp_status(response)
  
  # Try to extract error message from response body
  error_msg <- tryCatch({
    body <- httr2::resp_body_json(response)
    body$message %||% body$error %||% "Unknown error"
  }, error = function(e) {
    "Unable to parse error response"
  })
  
  # Format error based on status code
  if (status == 401) {
    stop(sprintf(
      "Authentication failed (401).\nEndpoint: %s\nPlease check your API key and workspace ID.",
      endpoint
    ), call. = FALSE)
  } else if (status == 403) {
    stop(sprintf(
      "Access denied (403).\nEndpoint: %s\nYou do not have permission to access this resource.\nDetails: %s",
      endpoint,
      error_msg
    ), call. = FALSE)
  } else if (status == 404) {
    stop(sprintf(
      "Resource not found (404).\nEndpoint: %s\nThis resource does not exist or you do not have access.\nDetails: %s",
      endpoint,
      error_msg
    ), call. = FALSE)
  } else if (status >= 500) {
    stop(sprintf(
      "Server error (%d).\nEndpoint: %s\nThe server encountered an error. Please try again later.\nDetails: %s",
      status,
      endpoint,
      error_msg
    ), call. = FALSE)
  } else {
    stop(sprintf(
      "API request failed (%d).\nEndpoint: %s\nDetails: %s",
      status,
      endpoint,
      error_msg
    ), call. = FALSE)
  }
}


#' Parse JSON Response
#'
#' Internal function to safely parse JSON response from API.
#'
#' @param response HTTP response object from httr2.
#'
#' @return Parsed JSON object.
#' @keywords internal
parse_json_response <- function(response) {
  tryCatch({
    httr2::resp_body_json(response, simplifyVector = FALSE)
  }, error = function(e) {
    stop(sprintf("Error parsing JSON response: %s", e$message), call. = FALSE)
  })
}


#' Format Error Message with Generic Access Denial
#'
#' Returns a generic error message that doesn't leak information about resource existence.
#' Used for schema/table access errors in accordance with security policy.
#'
#' @param resource_type Character. Type of resource (e.g., "schema", "table").
#' @param resource_name Character. Name of the resource (optional).
#'
#' @return Character. Generic error message.
#' @keywords internal
format_generic_access_error <- function(resource_type, resource_name = NULL) {
  if (!is.null(resource_name) && resource_name != "") {
    sprintf("This %s does not exist or you do not have access.", resource_type)
  } else {
    sprintf("This %s does not exist or you do not have access.", resource_type)
  }
}


#' Check if Response Indicates Access Denial
#'
#' Internal function to check if API response indicates access denial
#' (could be either non-existent or unauthorized).
#'
#' @param response HTTP response object from httr2.
#'
#' @return Logical. TRUE if access is denied.
#' @keywords internal
is_access_denied <- function(response) {
  status <- httr2::resp_status(response)
  return(status == 403 || status == 404)
}


#' Paginate Results
#'
#' Internal function to handle pagination of large result sets.
#'
#' @param data List. Full result set.
#' @param page Integer. Page number (0-indexed).
#' @param page_size Integer. Number of items per page.
#'
#' @return List with paginated data and metadata.
#' @keywords internal
paginate_results <- function(data, page = 0, page_size = 20) {
  total <- length(data)
  start_idx <- page * page_size + 1
  end_idx <- min((page + 1) * page_size, total)
  
  if (start_idx > total) {
    return(list(
      data = list(),
      page = page,
      page_size = page_size,
      total = total,
      pages = ceiling(total / page_size)
    ))
  }
  
  list(
    data = data[start_idx:end_idx],
    page = page,
    page_size = page_size,
    total = total,
    pages = ceiling(total / page_size)
  )
}

test_that("null-coalescing operator works correctly", {
  expect_equal(NULL %||% "default", "default")
  expect_equal("value" %||% "default", "value")
  expect_equal(0 %||% "default", 0)
  expect_equal("" %||% "default", "")
})

test_that("format_generic_access_error returns appropriate message", {
  msg1 <- cloudosR:::format_generic_access_error("schema", "public")
  expect_true(grepl("does not exist or you do not have access", msg1))
  
  msg2 <- cloudosR:::format_generic_access_error("table")
  expect_true(grepl("does not exist or you do not have access", msg2))
})


test_that("paginate_results handles edge cases", {
  data <- 1:100
  
  # First page
  result <- cloudosR:::paginate_results(data, page = 0, page_size = 20)
  expect_equal(length(result$data), 20)
  expect_equal(result$page, 0)
  expect_equal(result$total, 100)
  expect_equal(result$pages, 5)
  
  # Last page
  result <- cloudosR:::paginate_results(data, page = 4, page_size = 20)
  expect_equal(length(result$data), 20)
  expect_equal(result$page, 4)
  
  # Beyond last page
  result <- cloudosR:::paginate_results(data, page = 10, page_size = 20)
  expect_equal(length(result$data), 0)
})


test_that("build_query_string handles various inputs", {
  # Empty params
  expect_equal(cloudosR:::build_query_string(list()), "")
  
  # Single param
  result <- cloudosR:::build_query_string(list(param1 = "value1"))
  expect_true(grepl("param1=value1", result))
  
  # Multiple params
  result <- cloudosR:::build_query_string(list(param1 = "value1", param2 = "value2"))
  expect_true(grepl("param1=value1", result))
  expect_true(grepl("param2=value2", result))
  
  # NULL param should be skipped
  result <- cloudosR:::build_query_string(list(param1 = "value1", param2 = NULL))
  expect_true(grepl("param1=value1", result))
  expect_false(grepl("param2", result))
})

#' Validate SQL Query
#'
#' Validates SQL syntax and references before execution.
#'
#' @param profilename Character. Name of the configured profile to use. If empty or NULL, uses the default profile.
#' @param sql Character. SQL query to validate.
#'
#' @return List with validation results including isValid, tableReferences, and columnReferences.
#' @export
#'
#' @examples
#' \dontrun{
#'   # Validate a SQL query
#'   validation <- cloudos.sql_validate(
#'     profilename = "production",
#'     sql = "SELECT person_id FROM person WHERE year_of_birth >= 1960"
#'   )
#'   
#'   if (validation$isValid) {
#'     cat("SQL is valid\n")
#'     cat("Tables:", paste(validation$tableReferences, collapse = ", "), "\n")
#'   } else {
#'     cat("SQL is invalid:", validation$message, "\n")
#'   }
#' }
cloudos.sql_validate <- function(profilename = "", sql = "") {
  
  # Validate inputs
  validate_required_string(sql, "sql")
  
  # Load profile (uses default if profilename is empty)
  profile <- load_profile(profilename)
  
  # Build endpoint
  endpoint <- "/api/v2-cli/cohort-browser/sql-query/validate"
  
  # Build request body
  body <- list(
    sql = sql
  )
  
  # Build query params
  query_params <- list(
    teamId = profile$workspace_id
  )
  
  # Make request
  response <- tryCatch({
    http_post(profile, endpoint, body, query_params)
  }, error = function(e) {
    stop(sprintf(
      "Failed to validate SQL query:\n  Error: %s",
      e$message
    ), call. = FALSE)
  })
  
  # Return validation result
  return(response)
}


#' List Cohort Tables
#'
#' Retrieves the list of available database schemas and tables for a cohort.
#'
#' @param profilename Character. Name of the configured profile to use. If empty or NULL, uses the default profile.
#' @param cohort_id Character. ID of the cohort to query schemas for.
#'
#' @return List with schema information including databases, tables, and columns.
#' @export
#'
#' @examples
#' \dontrun{
#'   # Get available schemas for a cohort
#'   schemas <- cloudos.cohort_tables(
#'     profilename = "production",
#'     cohort_id = "your-cohort-id"
#'   )
#'   
#'   # Display databases
#'   cat("Available databases:\n")
#'   for (db in schemas) {
#'     cat("  -", db$database, "\n")
#'   }
#' }
cloudos.cohort_tables <- function(profilename = "", cohort_id = "") {
  
  # Validate inputs
  validate_required_string(cohort_id, "cohort_id")
  
  # Load profile (uses default if profilename is empty)
  profile <- load_profile(profilename)
  
  # Build endpoint
  endpoint <- "/api/v2-cli/cohort-browser/schemas"
  
  # Build query params
  query_params <- list(
    cohortId = cohort_id,
    teamId = profile$workspace_id
  )
  
  # Make request
  response <- tryCatch({
    http_get(profile, endpoint, query_params)
  }, error = function(e) {
    stop(sprintf(
      "Failed to fetch cohort schemas:\n  Error: %s",
      e$message
    ), call. = FALSE)
  })
  
  # Add class for custom printing
  class(response) <- c("cloudos_tables", class(response))
  attr(response, "cohort_id") <- cohort_id
  
  # Return schema information
  return(response)
}


#' Print method for cloudos_tables
#'
#' @param x A cloudos_tables object
#' @param ... Additional arguments (unused)
#' @return The \code{cloudos_tables} object \code{x}, returned invisibly.
#'   Called primarily for its side effect of printing a formatted list of
#'   schemas and tables to the console.
#' @export
print.cloudos_tables <- function(x, ...) {
  cohort_id <- attr(x, "cohort_id")
  schema_list <- x$schemas
  
  if (is.null(schema_list) || length(schema_list) == 0) {
    cat("No schemas found\n")
    return(invisible(x))
  }
  
  cat("Cohort", cohort_id, ":\n")
  
  for (schema in schema_list) {
    if (!is.null(schema$tables) && length(schema$tables) > 0) {
      for (table in schema$tables) {
        cat("  - ", schema$name, ".", table$name, "\n", sep = "")
        
        if (!is.null(table$columns) && length(table$columns) > 0) {
          # Show all columns
          for (col in table$columns) {
            cat("      - ", col$name, " (", col$dataType, ")\n", sep = "")
          }
        }
      }
    }
  }
  
  cat("\nTotal: ", length(schema_list), " database(s), ", 
      sum(sapply(schema_list, function(s) length(s$tables))), " table(s)\n", sep = "")
  cat("Use str() to see the full structure, or access x$schemas for the raw data\n")
  
  invisible(x)
}


#' Submit Async SQL Query
#'
#' Starts async SQL execution for a cohort and returns a task ID for tracking.
#'
#' @param profilename Character. Name of the configured profile to use. If empty or NULL, uses the default profile.
#' @param cohort_id Character. ID of the cohort to query.
#' @param sql Character. SQL query to execute.
#' @param pagination List (optional). Pagination settings with pageNumber and pageSize.
#'   Example: list(pageNumber = 0, pageSize = 100). If NULL, API returns default page.
#'
#' @return List with task metadata including task_id, status, and full response.
#' @export
#'
#' @examples
#' \dontrun{
#'   # Submit query without pagination
#'   task <- cloudos.query_submit_async(
#'     profilename = "production",
#'     cohort_id = "699edb4380a6867895f0c9e1",
#'     sql = "SELECT person_id FROM person LIMIT 100"
#'   )
#'   
#'   # Submit query with pagination for page 2 with 50 rows per page
#'   task <- cloudos.query_submit_async(
#'     profilename = "production",
#'     cohort_id = "699edb4380a6867895f0c9e1",
#'     sql = "SELECT person_id FROM person",
#'     pagination = list(pageNumber = 2, pageSize = 50)
#'   )
#'   print(task$task_id)
#' }
cloudos.query_submit_async <- function(profilename = "", cohort_id = "", sql = "", pagination = NULL) {
  
  # Validate inputs
  validate_required_string(cohort_id, "cohort_id")
  validate_required_string(sql, "sql")
  
  # Validate pagination if provided
  if (!is.null(pagination)) {
    if (!is.list(pagination)) {
      stop("Error: pagination must be a list with pageNumber and pageSize.", call. = FALSE)
    }
    if (is.null(pagination$pageNumber) || is.null(pagination$pageSize)) {
      stop("Error: pagination must contain both pageNumber and pageSize.", call. = FALSE)
    }
    if (!is.numeric(pagination$pageNumber) || pagination$pageNumber < 0) {
      stop("Error: pagination$pageNumber must be a non-negative integer.", call. = FALSE)
    }
    if (!is.numeric(pagination$pageSize) || pagination$pageSize < 1) {
      stop("Error: pagination$pageSize must be a positive integer.", call. = FALSE)
    }
  }
  
  # Load profile (uses default if profilename is empty)
  profile <- load_profile(profilename)
  
  # Build endpoint
  endpoint <- sprintf("/api/v2-cli/cohort-browser/cohort/%s/query-results/async", cohort_id)
  
  # Build request body
  body <- list(
    query = sql
  )
  
  # Add pagination if provided
  if (!is.null(pagination)) {
    body$pagination <- pagination
  }
  
  # Build query params
  query_params <- list(
    cohortId = cohort_id,
    teamId = profile$workspace_id
  )
  
  # Make request
  response <- tryCatch({
    http_post(profile, endpoint, body, query_params)
  }, error = function(e) {
    stop(sprintf(
      "Failed to submit query:\n  Cohort ID: %s\n  Error: %s",
      cohort_id,
      e$message
    ), call. = FALSE)
  })
  
  # Extract task information
  if (is.null(response$task) || is.null(response$task$`_id`)) {
    stop("Invalid response from server: missing task ID", call. = FALSE)
  }
  
  # Return task metadata
  result <- list(
    task_id = response$task$`_id`,
    status = response$task$status %||% "unknown",
    query = response$task$query %||% sql,
    type = response$task$type %||% "unknown",
    sync_execution_timeout = response$syncExecutionTimeout %||% 5000,
    full_response = response
  )
  
  message(sprintf("Query submitted successfully. Task ID: %s", result$task_id))
  
  return(result)
}


#' Check Async Query Status
#'
#' Returns the current status and metadata for a submitted async SQL task.
#'
#' @param profilename Character. Name of the configured profile to use. If empty or NULL, uses the default profile.
#' @param task_id Character. Task ID returned from cloudos.query_submit_async().
#'
#' @return List with task status, count of results, and other metadata.
#' @export
#'
#' @examples
#' \dontrun{
#'   status <- cloudos.query_status(
#'     profilename = "production",
#'     task_id = "69a5c58d626fe626da0025ce"
#'   )
#'   print(status$status)
#'   print(status$count_of_results)
#' }
cloudos.query_status <- function(profilename = "", task_id = "") {
  
  # Validate inputs
  validate_required_string(task_id, "task_id")
  
  # Load profile (uses default if profilename is empty)
  profile <- load_profile(profilename)
  
  # Build endpoint
  endpoint <- sprintf("/api/v2-cli/cohort-browser/async-tasks/%s", task_id)
  
  # Build query params
  query_params <- list(
    teamId = profile$workspace_id
  )
  
  # Make request
  response <- tryCatch({
    http_get(profile, endpoint, query_params)
  }, error = function(e) {
    stop(sprintf(
      "Failed to check task status:\n  Task ID: %s\n  Error: %s",
      task_id,
      e$message
    ), call. = FALSE)
  })
  
  # Extract status information
  result <- list(
    task_id = response$`_id` %||% task_id,
    status = response$status %||% "unknown",
    type = response$type %||% "unknown",
    count_of_results = response$countOfResults %||% 0,
    query = response$query %||% "",
    created_at = response$createdAt %||% "",
    started_at = response$startedAt %||% "",
    ended_at = response$endedAt %||% "",
    user = response$user %||% "",
    full_response = response
  )
  
  return(result)
}


#' Fetch Async Query Results
#'
#' Fetches results from a completed async SQL task and returns as a dataframe.
#' 
#' Note: Pagination is controlled when submitting the query via cloudos.query_submit_async(),
#' not when fetching results. This function returns whatever page the task was configured for.
#'
#' @param profilename Character. Name of the configured profile to use. If empty or NULL, uses the default profile.
#' @param task_id Character. Task ID returned from cloudos.query_submit_async().
#'
#' @return Data frame with query results from the page configured at submission time.
#' @export
#'
#' @examples
#' \dontrun{
#'   # Fetch results for a task (returns the page configured when query was submitted)
#'   results <- cloudos.query_results(
#'     profilename = "production",
#'     task_id = "69a5c58d626fe626da0025ce"
#'   )
#' }
cloudos.query_results <- function(profilename = "", task_id = "") {
  
  # Validate inputs
  validate_required_string(task_id, "task_id")
  
  # Load profile (uses default if profilename is empty)
  profile <- load_profile(profilename)
  
  # Build endpoint
  endpoint <- sprintf("/api/v2-cli/cohort-browser/async-tasks/%s/results", task_id)
  
  # Build query params
  query_params <- list(
    teamId = profile$workspace_id
  )
  
  # Fetch results
  response <- tryCatch({
    http_get(profile, endpoint, query_params)
  }, error = function(e) {
    stop(sprintf(
      "Failed to fetch results:\n  Task ID: %s\n  Error: %s",
      task_id,
      e$message
    ), call. = FALSE)
  })
  
  # Extract metadata
  # Note: API response structure is:
  #   - total: total number of rows across all pages
  #   - pageSize: total number of PAGES (not rows per page!)
  #   - pageNumber: current page index
  #   - data: array of rows for this page
  total_rows <- as.numeric(response$total %||% 0)
  total_pages <- as.numeric(response$pageSize %||% 1)  # This IS the number of pages
  current_page <- as.numeric(response$pageNumber %||% 0)
  rows_in_page <- length(response$data)
  
  # Extract column names
  column_names <- sapply(response$columns, function(col) {
    col$name %||% "unknown"
  })
  
  # Convert data to dataframe
  df <- convert_results_to_dataframe(response$data, column_names)
  
  # Add metadata as attributes
  attr(df, "total_rows") <- total_rows
  attr(df, "page") <- current_page
  attr(df, "page_size") <- rows_in_page  # Actual rows in this page
  attr(df, "total_pages") <- total_pages
  
  return(df)
}


#' Execute SQL Query (Orchestrator)
#'
#' High-level function that orchestrates the full query lifecycle:
#' submit -> poll status -> fetch results as dataframe.
#' 
#' IMPORTANT: Pagination works by submitting separate tasks for each page. When all_pages=TRUE,
#' this function submits multiple async tasks (one per page), waits for all to complete,
#' and combines the results. This may take longer for large result sets.
#'
#' @param profilename Character. Name of the configured profile to use. If empty or NULL, uses the default profile.
#' @param cohort_id Character. ID of the cohort to query.
#' @param sql Character. SQL query to execute.
#' @param poll_interval Integer. Seconds between status checks (default: 2).
#' @param max_wait Integer. Maximum seconds to wait for completion (default: 600).
#' @param page_size Integer. Number of rows per page (default: 1000).
#' @param all_pages Logical. Fetch all result pages automatically (default: TRUE).
#'   When TRUE, submits multiple async tasks (one per page) to fetch complete results.
#'
#' @return Data frame with query results.
#' @export
#'
#' @examples
#' \dontrun{
#'   # Fetch all results
#'   results <- cloudos.query(
#'     profilename = "production",
#'     cohort_id = "699edb4380a6867895f0c9e1",
#'     sql = "SELECT person_id, gender_concept_id FROM person LIMIT 500"
#'   )
#'   
#'   # Fetch only first page
#'   results_page1 <- cloudos.query(
#'     profilename = "production",
#'     cohort_id = "699edb4380a6867895f0c9e1",
#'     sql = "SELECT person_id FROM person",
#'     all_pages = FALSE,
#'     page_size = 100
#'   )
#' }
cloudos.query <- function(profilename = "",
                         cohort_id = "",
                         sql = "",
                         poll_interval = 2,
                         max_wait = 600,
                         page_size = 1000,
                         all_pages = TRUE) {
  
  # Validate inputs
  validate_required_string(cohort_id, "cohort_id")
  validate_required_string(sql, "sql")
  
  if (!is.numeric(poll_interval) || poll_interval < 1) {
    stop("Error: poll_interval must be at least 1 second.", call. = FALSE)
  }
  
  if (!is.numeric(max_wait) || max_wait < 1) {
    stop("Error: max_wait must be at least 1 second.", call. = FALSE)
  }
  
  if (!is.numeric(page_size) || page_size < 1) {
    stop("Error: page_size must be a positive integer.", call. = FALSE)
  }
  
  # Helper function to poll a task until completion
  poll_task <- function(task_id, task_description = "") {
    start_time <- Sys.time()
    
    while (TRUE) {
      elapsed <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
      
      if (elapsed >= max_wait) {
        stop(sprintf(
          "Task did not complete within %d seconds.\nTask ID: %s\nUse cloudos.query_status() to check progress.",
          max_wait,
          task_id
        ), call. = FALSE)
      }
      
      status_info <- tryCatch({
        cloudos.query_status(profilename, task_id)
      }, error = function(e) {
        stop(sprintf("Status check failed for %s: %s", task_description, e$message), call. = FALSE)
      })
      
      current_status <- tolower(trimws(status_info$status))
      
      if (current_status == "completed") {
        return(status_info)
      } else if (current_status == "failed") {
        stop(sprintf(
          "Query execution failed for %s.\nTask ID: %s\nCheck task status for details.",
          task_description,
          task_id
        ), call. = FALSE)
      } else if (current_status %in% c("pending", "running")) {
        message(sprintf("  %s: %s (%.1fs elapsed)...", task_description, trimws(status_info$status), elapsed))
      }
      
      Sys.sleep(poll_interval)
    }
  }
  
  # Step 1: Submit first query (page 0)
  message(sprintf("Submitting query%s...", if(all_pages) " for page 0" else ""))
  
  pagination <- list(pageNumber = 0, pageSize = page_size)
  
  task <- tryCatch({
    cloudos.query_submit_async(profilename, cohort_id, sql, pagination)
  }, error = function(e) {
    stop(sprintf("Query submission failed: %s", e$message), call. = FALSE)
  })
  
  task_id <- task$task_id
  
  # Step 2: Poll for completion
  message(sprintf("Polling for completion (max wait: %d seconds)...", max_wait))
  status_info <- poll_task(task_id, if(all_pages) "Page 0" else "Query")
  
  message(sprintf(
    "Page 0 completed successfully (%s results)",
    status_info$count_of_results
  ))
  
  # Step 3: Fetch first page results
  message("Fetching page 0 results...")
  first_page <- tryCatch({
    cloudos.query_results(profilename, task_id)
  }, error = function(e) {
    stop(sprintf("Failed to fetch results: %s", e$message), call. = FALSE)
  })
  
  # If not fetching all pages, return first page
  if (!all_pages) {
    total_rows <- attr(first_page, "total_rows")
    total_pages <- attr(first_page, "total_pages")
    
    if (total_pages > 1) {
      message(sprintf(
        "\nNote: Query has %d total rows across %d pages. Only page 0 (%d rows) returned.",
        total_rows,
        total_pages,
        nrow(first_page)
      ))
      message("Use all_pages=TRUE to fetch all results.")
    }
    
    message("Query complete!")
    return(first_page)
  }
  
  # Check if there are more pages
  total_rows <- attr(first_page, "total_rows")
  total_pages <- attr(first_page, "total_pages")
  
  if (total_pages <= 1) {
    message(sprintf("All results fetched: %d rows", nrow(first_page)))
    message("Query complete!")
    return(first_page)
  }
  
  # Step 4: Submit tasks for remaining pages
  message(sprintf(
    "\nFetching remaining pages (1 to %d) - total: %d rows across %d pages...",
    total_pages - 1,
    total_rows,
    total_pages
  ))
  
  remaining_tasks <- list()
  
  for (page_num in 1:(total_pages - 1)) {
    pagination <- list(pageNumber = page_num, pageSize = page_size)
    
    task <- tryCatch({
      cloudos.query_submit_async(profilename, cohort_id, sql, pagination)
    }, error = function(e) {
      stop(sprintf("Failed to submit query for page %d: %s", page_num, e$message), call. = FALSE)
    })
    
    remaining_tasks[[page_num]] <- list(
      task_id = task$task_id,
      page_num = page_num
    )
    
    message(sprintf("  Submitted page %d (task ID: %s)", page_num, task$task_id))
    
    # Small delay to avoid overwhelming the server
    Sys.sleep(0.2)
  }
  
  # Step 5: Poll all remaining tasks
  message("\nWaiting for all pages to complete...")
  
  page_results <- list(first_page)
  
  for (task_info in remaining_tasks) {
    page_num <- task_info$page_num
    task_id <- task_info$task_id
    
    message(sprintf("Polling page %d...", page_num))
    status_info <- poll_task(task_id, sprintf("Page %d", page_num))
    
    message(sprintf("Page %d completed, fetching results...", page_num))
    
    page_df <- tryCatch({
      cloudos.query_results(profilename, task_id)
    }, error = function(e) {
      stop(sprintf("Failed to fetch page %d results: %s", page_num, e$message), call. = FALSE)
    })
    
    page_results[[page_num + 1]] <- page_df
  }
  
  # Step 6: Combine all pages
  message("\nCombining all pages...")
  combined_df <- do.call(rbind, page_results)
  
  # Update metadata
  attr(combined_df, "total_rows") <- total_rows
  attr(combined_df, "page") <- 0
  attr(combined_df, "page_size") <- total_rows
  attr(combined_df, "total_pages") <- 1
  attr(combined_df, "all_pages_fetched") <- TRUE
  
  message(sprintf("All pages fetched: %d rows", nrow(combined_df)))
  message("Query complete!")
  
  return(combined_df)
}


#' Convert Query Results to Dataframe
#'
#' Internal function to convert API response data to a dataframe.
#'
#' @param data List. Data rows from API response.
#' @param column_names Character vector. Column names.
#'
#' @return Data frame with query results.
#' @keywords internal
convert_results_to_dataframe <- function(data, column_names) {
  
  # Handle empty results
  if (is.null(data) || length(data) == 0) {
    # Create empty dataframe with correct columns
    df <- as.data.frame(matrix(ncol = length(column_names), nrow = 0))
    colnames(df) <- column_names
    return(df)
  }
  
  # Convert each row to a named list
  rows <- lapply(data, function(row) {
    # Create a named list for the row
    row_data <- list()
    for (i in seq_along(column_names)) {
      col_name <- column_names[i]
      # API returns rows as objects with column names as keys
      value <- row[[col_name]]
      # Handle NULL values
      if (is.null(value)) {
        row_data[[col_name]] <- NA
      } else {
        row_data[[col_name]] <- value
      }
    }
    return(row_data)
  })
  
  # Convert to dataframe
  df <- as.data.frame(do.call(rbind, lapply(rows, function(r) {
    as.data.frame(r, stringsAsFactors = FALSE)
  })), stringsAsFactors = FALSE)
  
  # Ensure column names are set correctly
  colnames(df) <- column_names
  
  return(df)
}

test_that("convert_results_to_dataframe handles empty results", {
  result <- cloudosR:::convert_results_to_dataframe(list(), c("col1", "col2"))
  
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 0)
  expect_equal(ncol(result), 2)
  expect_equal(colnames(result), c("col1", "col2"))
})

test_that("convert_results_to_dataframe handles NULL values", {
  data <- list(
    list(col1 = "value1", col2 = NULL),
    list(col1 = NULL, col2 = "value2")
  )
  
  result <- cloudosR:::convert_results_to_dataframe(data, c("col1", "col2"))
  
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 2)
  expect_true(is.na(result[1, "col2"]))
  expect_true(is.na(result[2, "col1"]))
})


test_that("cloudos.query_submit_async validates inputs", {
  skip_on_cran()
  
  # Create a temporary config with a profile
  temp_config <- tempfile("cloudos_test_config")
  dir.create(temp_config, recursive = TRUE)
  old_env <- Sys.getenv("CLOUDOS_CONFIG_DIR", unset = NA)
  Sys.setenv(CLOUDOS_CONFIG_DIR = temp_config)
  
  on.exit({
    if (is.na(old_env)) {
      Sys.unsetenv("CLOUDOS_CONFIG_DIR")
    } else {
      Sys.setenv(CLOUDOS_CONFIG_DIR = old_env)
    }
    unlink(temp_config, recursive = TRUE)
  })
  
  # Create a test profile
  cloudos.configure(
    profilename = "test",
    apikey = "test-key",
    workspace_id = "test-workspace",
    set_default = TRUE
  )
  
  expect_error(
    cloudos.query_submit_async(profilename = "test", cohort_id = "", sql = "SELECT 1"),
    "cohort_id is required"
  )
  
  expect_error(
    cloudos.query_submit_async(profilename = "test", cohort_id = "123", sql = ""),
    "sql is required"
  )
})


test_that("cloudos.query validates poll_interval and max_wait", {
  skip_on_cran()
  
  # Create a temporary config with a profile
  temp_config <- tempfile("cloudos_test_config")
  dir.create(temp_config, recursive = TRUE)
  old_env <- Sys.getenv("CLOUDOS_CONFIG_DIR", unset = NA)
  Sys.setenv(CLOUDOS_CONFIG_DIR = temp_config)
  
  on.exit({
    if (is.na(old_env)) {
      Sys.unsetenv("CLOUDOS_CONFIG_DIR")
    } else {
      Sys.setenv(CLOUDOS_CONFIG_DIR = old_env)
    }
    unlink(temp_config, recursive = TRUE)
  })
  
  # Create a test profile
  cloudos.configure(
    profilename = "test",
    apikey = "test-key",
    workspace_id = "test-workspace",
    set_default = TRUE
  )
  
  expect_error(
    cloudos.query(
      profilename = "test",
      cohort_id = "123",
      sql = "SELECT 1",
      poll_interval = 0
    ),
    "poll_interval must be at least 1 second"
  )
  
  expect_error(
    cloudos.query(
      profilename = "test",
      cohort_id = "123",
      sql = "SELECT 1",
      max_wait = 0
    ),
    "max_wait must be at least 1 second"
  )
})

#' Make HTTP GET Request
#'
#' Internal function to make authenticated GET requests to Lifebit Platform API.
#'
#' @param profile List. Profile configuration from load_profile().
#' @param endpoint Character. API endpoint path (without base URL).
#' @param query_params List. Query parameters to include in request.
#'
#' @return Parsed JSON response.
#' @keywords internal
http_get <- function(profile, endpoint, query_params = list()) {
  
  # Build full URL
  base_url <- profile$base_url
  url <- paste0(base_url, endpoint)
  
  # Add teamId to query params if not present
  if (!"teamId" %in% names(query_params)) {
    query_params$teamId <- profile$workspace_id
  }
  
  # Create request with specific options to avoid hanging
  req <- httr2::request(url) |>
    httr2::req_url_query(!!!query_params) |>
    httr2::req_headers(
      "apikey" = profile$apikey,
      "Content-Type" = "application/json",
      "Accept" = "application/json"
    ) |>
    httr2::req_timeout(60) |>
    httr2::req_options(
      connecttimeout = 10,
      tcp_keepalive = 1L,
      tcp_keepidle = 120L,
      tcp_keepintvl = 60L
    ) |>
    httr2::req_retry(max_tries = 1)  # No retries to avoid hanging
  
  # Perform request
  response <- tryCatch({
    httr2::req_perform(req)
  }, error = function(e) {
    stop(sprintf("HTTP request failed: %s", e$message), call. = FALSE)
  })
  
  # Check status
  status <- httr2::resp_status(response)
  if (!(status %in% c(200, 202))) {
    handle_api_error(response, endpoint)
  }
  
  # Parse and return response
  return(parse_json_response(response))
}


#' Make HTTP POST Request
#'
#' Internal function to make authenticated POST requests to Lifebit Platform API.
#'
#' @param profile List. Profile configuration from load_profile().
#' @param endpoint Character. API endpoint path (without base URL).
#' @param body List. Request body to send as JSON.
#' @param query_params List. Query parameters to include in request.
#'
#' @return Parsed JSON response.
#' @keywords internal
http_post <- function(profile, endpoint, body = list(), query_params = list()) {
  
  # Build full URL
  base_url <- profile$base_url
  url <- paste0(base_url, endpoint)
  
  # Add teamId to query params if not present
  if (!"teamId" %in% names(query_params)) {
    query_params$teamId <- profile$workspace_id
  }
  
  # Create request with specific options to avoid hanging
  req <- httr2::request(url) |>
    httr2::req_url_query(!!!query_params) |>
    httr2::req_headers(
      "apikey" = profile$apikey,
      "Content-Type" = "application/json",
      "Accept" = "application/json"
    ) |>
    httr2::req_body_json(body) |>
    httr2::req_timeout(60) |>
    httr2::req_options(
      connecttimeout = 10,
      tcp_keepalive = 1L,
      tcp_keepidle = 120L,
      tcp_keepintvl = 60L
    ) |>
    httr2::req_retry(max_tries = 1)  # No retries to avoid hanging
  
  # Perform request
  response <- tryCatch({
    httr2::req_perform(req)
  }, error = function(e) {
    stop(sprintf("HTTP request failed: %s", e$message), call. = FALSE)
  })
  
  # Check status
  status <- httr2::resp_status(response)
  if (!(status %in% c(200, 202))) {
    handle_api_error(response, endpoint)
  }
  
  # Parse and return response
  return(parse_json_response(response))
}


#' Make HTTP Request with Retry
#'
#' Internal function to make HTTP requests with retry logic for transient failures.
#'
#' @param method Character. HTTP method ("GET" or "POST").
#' @param profile List. Profile configuration from load_profile().
#' @param endpoint Character. API endpoint path (without base URL).
#' @param body List. Request body (for POST requests).
#' @param query_params List. Query parameters.
#' @param max_retries Integer. Maximum number of retries (default: 3).
#' @param retry_delay Integer. Delay between retries in seconds (default: 2).
#'
#' @return Parsed JSON response.
#' @keywords internal
http_request_with_retry <- function(method, 
                                    profile, 
                                    endpoint, 
                                    body = list(), 
                                    query_params = list(),
                                    max_retries = 3,
                                    retry_delay = 2) {
  
  last_error <- NULL
  
  for (attempt in 1:max_retries) {
    tryCatch({
      if (method == "GET") {
        return(http_get(profile, endpoint, query_params))
      } else if (method == "POST") {
        return(http_post(profile, endpoint, body, query_params))
      } else {
        stop(sprintf("Unsupported HTTP method: %s", method))
      }
    }, error = function(e) {
      last_error <<- e
      
      # Don't retry on authentication or authorization errors
      if (grepl("401|403", e$message)) {
        stop(e)
      }
      
      # Don't retry on client errors (4xx except 408)
      if (grepl("40[024-9]", e$message)) {
        stop(e)
      }
      
      # Retry on server errors (5xx) or network errors
      if (attempt < max_retries) {
        message(sprintf(
          "Request failed (attempt %d/%d): %s\nRetrying in %d seconds...",
          attempt,
          max_retries,
          e$message,
          retry_delay
        ))
        Sys.sleep(retry_delay)
      }
    })
  }
  
  # If we get here, all retries failed
  stop(sprintf(
    "Request failed after %d attempts. Last error: %s",
    max_retries,
    last_error$message
  ), call. = FALSE)
}


#' Build Query String
#'
#' Internal function to build URL query string from list of parameters.
#'
#' @param params List. Named list of query parameters.
#'
#' @return Character. URL-encoded query string.
#' @keywords internal
build_query_string <- function(params) {
  if (length(params) == 0) {
    return("")
  }
  
  # Filter out NULL and empty values
  params <- params[!sapply(params, is.null)]
  params <- params[sapply(params, function(x) length(x) > 0 && x != "")]
  
  if (length(params) == 0) {
    return("")
  }
  
  # Build query string
  pairs <- mapply(function(name, value) {
    if (length(value) > 1) {
      # Handle array parameters (e.g., datasourceIds[])
      paste(sapply(value, function(v) {
        sprintf("%s=%s", utils::URLencode(name, reserved = TRUE), utils::URLencode(as.character(v), reserved = TRUE))
      }), collapse = "&")
    } else {
      sprintf("%s=%s", utils::URLencode(name, reserved = TRUE), utils::URLencode(as.character(value), reserved = TRUE))
    }
  }, names(params), params, SIMPLIFY = FALSE, USE.NAMES = FALSE)
  
  paste0("?", paste(unlist(pairs), collapse = "&"))
}

---
title: "Getting Started with cloudosR"
author: "Leila Mansouri <leila.mansouri@lifebit.ai>"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with cloudosR}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Introduction

The `cloudosR` package provides an R client for interacting with the Lifebit Platform. It enables you to:

- Configure and manage Lifebit Platform profiles with credentials
- Execute read-only SQL queries on cohort data
- Retrieve results as R dataframes for analysis

## Installation

```{r eval=FALSE}
# Install from CRAN (recommended)
install.packages("cloudosR")

# Or install development version from GitHub
devtools::install_github("lifebit-ai/cloudosR")

# Or install from local source
install.packages("path/to/cloudosR", repos = NULL, type = "source")
```

## Configuration

Before using any API functions, you need to configure a profile with your Lifebit Platform credentials.

### Setting up a Profile

```{r eval=FALSE}
library(cloudosR)

# Configure a profile with your credentials
cloudos.configure(
  profilename = "production",
  apikey = "your-api-key-here",
  workspace_id = "your-workspace-id",
  base_url = "https://cloudos.lifebit.ai",  # Optional, this is the default
  set_default = TRUE  # Set as default profile
)
```

The configuration is stored securely in your R user config directory with restricted permissions (0600). You can find the location with `tools::R_user_dir("cloudosR", "config")`.

### Using Default Profiles

When you set `set_default = TRUE`, you don't need to specify the `profilename` parameter in subsequent function calls:

```{r eval=FALSE}
# With default profile set, you can omit profilename
results <- cloudos.query(
  cohort_id = "1a2b3c4d5e6f7g8h9i10j11k",
  sql = "SELECT * FROM table LIMIT 10"
)
```

### Managing Multiple Profiles

You can configure multiple profiles for different environments:

```{r eval=FALSE}
# Production profile
cloudos.configure(
  profilename = "production",
  apikey = "prod-api-key",
  workspace_id = "prod-workspace-id",
  set_default = TRUE
)

# Staging profile
cloudos.configure(
  profilename = "staging",
  apikey = "staging-api-key",
  workspace_id = "staging-workspace-id"
)

# List all configured profiles
profiles <- cloudos.profile_list()
print(profiles)
```

## Querying Data

The package provides both high-level and low-level functions for executing SQL queries.

### High-Level Query Execution (Recommended)

The `cloudos.query()` function handles the entire query lifecycle automatically:

```{r eval=FALSE}
# Simple query (uses default profile)
results <- cloudos.query(
  cohort_id = "1a2b3c4d5e6f7g8h9i10j11k",
  sql = "SELECT person_id, gender_concept_id, birth_datetime 
         FROM person 
         LIMIT 10"
)

# View results
head(results)
str(results)

# Access metadata
attr(results, "total_rows")
attr(results, "total_pages")
```

### Low-Level Query Control

For more control over the query process, you can use the individual steps:

```{r eval=FALSE}
# Step 1: Submit query with pagination
task <- cloudos.query_submit_async(
  profilename = "production",
  cohort_id = "1a2b3c4d5e6f7g8h9i10j11k",
  sql = "SELECT * FROM table LIMIT 10",
  pagination = list(pageNumber = 0, pageSize = 100)  # Optional
)

task_id <- task$task_id
print(task_id)

# Step 2: Check status
status <- cloudos.query_status(
  profilename = "production",
  task_id = task_id
)

print(status$status)  # "pending", "running", "completed", or "failed"
print(status$count_of_results)

# Step 3: Fetch results when completed
results <- cloudos.query_results(
  profilename = "production",
  task_id = task_id
)
```

### Working with Large Result Sets

The package automatically handles pagination by submitting multiple async tasks:

```{r eval=FALSE}
# Fetch all pages automatically (default)
# Note: This submits separate tasks for each page
all_results <- cloudos.query(
  cohort_id = "1a2b3c4d5e6f7g8h9i10j11k",
  sql = "SELECT * FROM person",
  page_size = 1000,    # Rows per page (default: 1000)
  all_pages = TRUE     # Fetch all pages (default: TRUE)
)

# This will:
# 1. Submit query for page 0
# 2. Wait for completion and fetch results
# 3. Calculate total pages from response metadata
# 4. Submit separate tasks for pages 1, 2, 3...
# 5. Wait for all tasks to complete
# 6. Combine and return all results

# Fetch only first page
first_page <- cloudos.query(
  cohort_id = "1a2b3c4d5e6f7g8h9i10j11k",
  sql = "SELECT * FROM person",
  page_size = 100,
  all_pages = FALSE    # Only fetch first page
)
```

## Advanced Usage

### Custom Polling Parameters

You can customize how the package polls for query completion:

```{r eval=FALSE}
results <- cloudos.query(
  cohort_id = "1a2b3c4d5e6f7g8h9i10j11k",
  sql = "SELECT COUNT(*) FROM person",
  poll_interval = 5,    # Check every 5 seconds (default: 2)
  max_wait = 900        # Wait up to 15 minutes (default: 600)
)
```

### Complex SQL Queries

The package supports any valid SQL query that your cohort allows:

```{r eval=FALSE}
# Join multiple tables
sql_query <- "
  SELECT 
    p.person_id,
    p.gender_concept_id,
    c.condition_concept_id,
    c.condition_start_date
  FROM person p
  JOIN condition_occurrence c ON p.person_id = c.person_id
  WHERE p.birth_datetime > '1980-01-01'
  LIMIT 1000
"

results <- cloudos.query(
  cohort_id = "1a2b3c4d5e6f7g8h9i10j11k",
  sql = sql_query
)
```

## Security Best Practices

1. **Never commit credentials**: Add `.cloudos_config.json` to your `.gitignore` file
2. **Use environment-specific profiles**: Separate production and development credentials
3. **Rotate API keys regularly**: Update profiles when keys change
4. **Restrict file permissions**: The package automatically sets config file permissions to 0600

## Error Handling

The package provides informative error messages:

```{r eval=FALSE}
# Authentication errors
tryCatch({
  cloudos.query(
    profilename = "invalid_profile",
    cohort_id = "123",
    sql = "SELECT 1"
  )
}, error = function(e) {
  message("Error caught: ", e$message)
})

# Query timeout
tryCatch({
  cloudos.query(
    cohort_id = "123",
    sql = "SELECT * FROM large_table",
    max_wait = 5  # Very short timeout
  )
}, error = function(e) {
  message("Query timed out: ", e$message)
})
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/query.R
\name{cloudos.query}
\alias{cloudos.query}
\title{Execute SQL Query (Orchestrator)}
\usage{
cloudos.query(
  profilename = "",
  cohort_id = "",
  sql = "",
  poll_interval = 2,
  max_wait = 600,
  page_size = 1000,
  all_pages = TRUE
)
}
\arguments{
\item{profilename}{Character. Name of the configured profile to use. If empty or NULL, uses the default profile.}

\item{cohort_id}{Character. ID of the cohort to query.}

\item{sql}{Character. SQL query to execute.}

\item{poll_interval}{Integer. Seconds between status checks (default: 2).}

\item{max_wait}{Integer. Maximum seconds to wait for completion (default: 600).}

\item{page_size}{Integer. Number of rows per page (default: 1000).}

\item{all_pages}{Logical. Fetch all result pages automatically (default: TRUE).
When TRUE, submits multiple async tasks (one per page) to fetch complete results.}
}
\value{
Data frame with query results.
}
\description{
High-level function that orchestrates the full query lifecycle:
submit -> poll status -> fetch results as dataframe.
}
\details{
IMPORTANT: Pagination works by submitting separate tasks for each page. When all_pages=TRUE,
this function submits multiple async tasks (one per page), waits for all to complete,
and combines the results. This may take longer for large result sets.
}
\examples{
\dontrun{
  # Fetch all results
  results <- cloudos.query(
    profilename = "production",
    cohort_id = "699edb4380a6867895f0c9e1",
    sql = "SELECT person_id, gender_concept_id FROM person LIMIT 500"
  )
  
  # Fetch only first page
  results_page1 <- cloudos.query(
    profilename = "production",
    cohort_id = "699edb4380a6867895f0c9e1",
    sql = "SELECT person_id FROM person",
    all_pages = FALSE,
    page_size = 100
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{format_generic_access_error}
\alias{format_generic_access_error}
\title{Format Error Message with Generic Access Denial}
\usage{
format_generic_access_error(resource_type, resource_name = NULL)
}
\arguments{
\item{resource_type}{Character. Type of resource (e.g., "schema", "table").}

\item{resource_name}{Character. Name of the resource (optional).}
}
\value{
Character. Generic error message.
}
\description{
Returns a generic error message that doesn't leak information about resource existence.
Used for schema/table access errors in accordance with security policy.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{cloudos.configure}
\alias{cloudos.configure}
\title{Configure Lifebit Platform Profile}
\usage{
cloudos.configure(
  profilename = "",
  apikey = "",
  workspace_id = "",
  base_url = "https://cloudos.lifebit.ai",
  set_default = FALSE
)
}
\arguments{
\item{profilename}{Character. Name of the profile to create or update.}

\item{apikey}{Character. API key for authentication.}

\item{workspace_id}{Character. Workspace/team ID for API requests.}

\item{base_url}{Character. Base URL for Lifebit Platform API (default: "https://cloudos.lifebit.ai").}

\item{set_default}{Logical. If TRUE, sets this profile as the default (default: FALSE).}
}
\value{
Invisible NULL. Prints a success message.
}
\description{
Stores API credentials and workspace context for a named profile.
This is the required first step before any API wrapper call.
}
\examples{
\dontrun{
  cloudos.configure(
    profilename = "production",
    apikey = "your-api-key",
    workspace_id = "5c6d3e9bd954e800b23f8c62",
    set_default = TRUE
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{parse_json_response}
\alias{parse_json_response}
\title{Parse JSON Response}
\usage{
parse_json_response(response)
}
\arguments{
\item{response}{HTTP response object from httr2.}
}
\value{
Parsed JSON object.
}
\description{
Internal function to safely parse JSON response from API.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/query.R
\name{cloudos.query_submit_async}
\alias{cloudos.query_submit_async}
\title{Submit Async SQL Query}
\usage{
cloudos.query_submit_async(
  profilename = "",
  cohort_id = "",
  sql = "",
  pagination = NULL
)
}
\arguments{
\item{profilename}{Character. Name of the configured profile to use. If empty or NULL, uses the default profile.}

\item{cohort_id}{Character. ID of the cohort to query.}

\item{sql}{Character. SQL query to execute.}

\item{pagination}{List (optional). Pagination settings with pageNumber and pageSize.
Example: list(pageNumber = 0, pageSize = 100). If NULL, API returns default page.}
}
\value{
List with task metadata including task_id, status, and full response.
}
\description{
Starts async SQL execution for a cohort and returns a task ID for tracking.
}
\examples{
\dontrun{
  # Submit query without pagination
  task <- cloudos.query_submit_async(
    profilename = "production",
    cohort_id = "699edb4380a6867895f0c9e1",
    sql = "SELECT person_id FROM person LIMIT 100"
  )
  
  # Submit query with pagination for page 2 with 50 rows per page
  task <- cloudos.query_submit_async(
    profilename = "production",
    cohort_id = "699edb4380a6867895f0c9e1",
    sql = "SELECT person_id FROM person",
    pagination = list(pageNumber = 2, pageSize = 50)
  )
  print(task$task_id)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/query.R
\name{cloudos.query_status}
\alias{cloudos.query_status}
\title{Check Async Query Status}
\usage{
cloudos.query_status(profilename = "", task_id = "")
}
\arguments{
\item{profilename}{Character. Name of the configured profile to use. If empty or NULL, uses the default profile.}

\item{task_id}{Character. Task ID returned from cloudos.query_submit_async().}
}
\value{
List with task status, count of results, and other metadata.
}
\description{
Returns the current status and metadata for a submitted async SQL task.
}
\examples{
\dontrun{
  status <- cloudos.query_status(
    profilename = "production",
    task_id = "69a5c58d626fe626da0025ce"
  )
  print(status$status)
  print(status$count_of_results)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/query.R
\name{cloudos.sql_validate}
\alias{cloudos.sql_validate}
\title{Validate SQL Query}
\usage{
cloudos.sql_validate(profilename = "", sql = "")
}
\arguments{
\item{profilename}{Character. Name of the configured profile to use. If empty or NULL, uses the default profile.}

\item{sql}{Character. SQL query to validate.}
}
\value{
List with validation results including isValid, tableReferences, and columnReferences.
}
\description{
Validates SQL syntax and references before execution.
}
\examples{
\dontrun{
  # Validate a SQL query
  validation <- cloudos.sql_validate(
    profilename = "production",
    sql = "SELECT person_id FROM person WHERE year_of_birth >= 1960"
  )
  
  if (validation$isValid) {
    cat("SQL is valid\n")
    cat("Tables:", paste(validation$tableReferences, collapse = ", "), "\n")
  } else {
    cat("SQL is invalid:", validation$message, "\n")
  }
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/query.R
\name{cloudos.query_results}
\alias{cloudos.query_results}
\title{Fetch Async Query Results}
\usage{
cloudos.query_results(profilename = "", task_id = "")
}
\arguments{
\item{profilename}{Character. Name of the configured profile to use. If empty or NULL, uses the default profile.}

\item{task_id}{Character. Task ID returned from cloudos.query_submit_async().}
}
\value{
Data frame with query results from the page configured at submission time.
}
\description{
Fetches results from a completed async SQL task and returns as a dataframe.
}
\details{
Note: Pagination is controlled when submitting the query via cloudos.query_submit_async(),
not when fetching results. This function returns whatever page the task was configured for.
}
\examples{
\dontrun{
  # Fetch results for a task (returns the page configured when query was submitted)
  results <- cloudos.query_results(
    profilename = "production",
    task_id = "69a5c58d626fe626da0025ce"
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{null-coalesce}
\alias{null-coalesce}
\alias{\%||\%}
\title{Null-coalescing Operator}
\usage{
x \%||\% y
}
\arguments{
\item{x}{First value to check.}

\item{y}{Default value if x is NULL.}
}
\value{
x if not NULL, otherwise y.
}
\description{
Returns the left-hand side if it's not NULL, otherwise returns the right-hand side.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/query.R
\name{print.cloudos_tables}
\alias{print.cloudos_tables}
\title{Print method for cloudos_tables}
\usage{
\method{print}{cloudos_tables}(x, ...)
}
\arguments{
\item{x}{A cloudos_tables object}

\item{...}{Additional arguments (unused)}
}
\value{
The \code{cloudos_tables} object \code{x}, returned invisibly.
  Called primarily for its side effect of printing a formatted list of
  schemas and tables to the console.
}
\description{
Print method for cloudos_tables
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/http.R
\name{http_request_with_retry}
\alias{http_request_with_retry}
\title{Make HTTP Request with Retry}
\usage{
http_request_with_retry(
  method,
  profile,
  endpoint,
  body = list(),
  query_params = list(),
  max_retries = 3,
  retry_delay = 2
)
}
\arguments{
\item{method}{Character. HTTP method ("GET" or "POST").}

\item{profile}{List. Profile configuration from load_profile().}

\item{endpoint}{Character. API endpoint path (without base URL).}

\item{body}{List. Request body (for POST requests).}

\item{query_params}{List. Query parameters.}

\item{max_retries}{Integer. Maximum number of retries (default: 3).}

\item{retry_delay}{Integer. Delay between retries in seconds (default: 2).}
}
\value{
Parsed JSON response.
}
\description{
Internal function to make HTTP requests with retry logic for transient failures.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{cloudos.profile_list}
\alias{cloudos.profile_list}
\title{List Lifebit Platform Profiles}
\usage{
cloudos.profile_list()
}
\value{
Data frame with profile names and metadata (workspace_id, default, created_at, updated_at).
        Returns empty data frame if no profiles are configured.
}
\description{
Lists configured profiles so users can confirm available environments.
}
\examples{
\dontrun{
  profiles <- cloudos.profile_list()
  print(profiles)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{handle_api_error}
\alias{handle_api_error}
\title{Handle API Error Response}
\usage{
handle_api_error(response, endpoint)
}
\arguments{
\item{response}{HTTP response object from httr.}

\item{endpoint}{Character. Endpoint that was called (for error context).}
}
\value{
NULL (throws error with formatted message).
}
\description{
Internal function to handle API error responses and generate user-friendly error messages.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/http.R
\name{http_post}
\alias{http_post}
\title{Make HTTP POST Request}
\usage{
http_post(profile, endpoint, body = list(), query_params = list())
}
\arguments{
\item{profile}{List. Profile configuration from load_profile().}

\item{endpoint}{Character. API endpoint path (without base URL).}

\item{body}{List. Request body to send as JSON.}

\item{query_params}{List. Query parameters to include in request.}
}
\value{
Parsed JSON response.
}
\description{
Internal function to make authenticated POST requests to Lifebit Platform API.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{validate_required_string}
\alias{validate_required_string}
\title{Validate Required String Parameter}
\usage{
validate_required_string(value, param_name)
}
\arguments{
\item{value}{Character. Value to validate.}

\item{param_name}{Character. Name of the parameter (for error messages).}
}
\value{
NULL (throws error if validation fails).
}
\description{
Internal function to validate that a string parameter is not empty or NULL.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/query.R
\name{cloudos.cohort_tables}
\alias{cloudos.cohort_tables}
\title{List Cohort Tables}
\usage{
cloudos.cohort_tables(profilename = "", cohort_id = "")
}
\arguments{
\item{profilename}{Character. Name of the configured profile to use. If empty or NULL, uses the default profile.}

\item{cohort_id}{Character. ID of the cohort to query schemas for.}
}
\value{
List with schema information including databases, tables, and columns.
}
\description{
Retrieves the list of available database schemas and tables for a cohort.
}
\examples{
\dontrun{
  # Get available schemas for a cohort
  schemas <- cloudos.cohort_tables(
    profilename = "production",
    cohort_id = "your-cohort-id"
  )
  
  # Display databases
  cat("Available databases:\n")
  for (db in schemas) {
    cat("  -", db$database, "\n")
  }
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{paginate_results}
\alias{paginate_results}
\title{Paginate Results}
\usage{
paginate_results(data, page = 0, page_size = 20)
}
\arguments{
\item{data}{List. Full result set.}

\item{page}{Integer. Page number (0-indexed).}

\item{page_size}{Integer. Number of items per page.}
}
\value{
List with paginated data and metadata.
}
\description{
Internal function to handle pagination of large result sets.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/http.R
\name{build_query_string}
\alias{build_query_string}
\title{Build Query String}
\usage{
build_query_string(params)
}
\arguments{
\item{params}{List. Named list of query parameters.}
}
\value{
Character. URL-encoded query string.
}
\description{
Internal function to build URL query string from list of parameters.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{load_profile}
\alias{load_profile}
\title{Load Profile Configuration}
\usage{
load_profile(profilename = "")
}
\arguments{
\item{profilename}{Character. Name of the profile to load. If empty, loads default profile.}
}
\value{
List with profile configuration (apikey, workspace_id, base_url).
}
\description{
Internal function to load a specific profile's configuration.
If profilename is empty or NULL, loads the default profile.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{get_config_dir}
\alias{get_config_dir}
\title{Get Config Directory Path}
\usage{
get_config_dir()
}
\value{
Character. Path to config directory.
}
\description{
Internal function to get the Lifebit Platform config directory path.
Uses tools::R_user_dir() for CRAN-compliant persistent storage.
Can be overridden with CLOUDOS_CONFIG_DIR environment variable.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/query.R
\name{convert_results_to_dataframe}
\alias{convert_results_to_dataframe}
\title{Convert Query Results to Dataframe}
\usage{
convert_results_to_dataframe(data, column_names)
}
\arguments{
\item{data}{List. Data rows from API response.}

\item{column_names}{Character vector. Column names.}
}
\value{
Data frame with query results.
}
\description{
Internal function to convert API response data to a dataframe.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{is_access_denied}
\alias{is_access_denied}
\title{Check if Response Indicates Access Denial}
\usage{
is_access_denied(response)
}
\arguments{
\item{response}{HTTP response object from httr2.}
}
\value{
Logical. TRUE if access is denied.
}
\description{
Internal function to check if API response indicates access denial
(could be either non-existent or unauthorized).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{get_config_file}
\alias{get_config_file}
\title{Get Config File Path}
\usage{
get_config_file()
}
\value{
Character. Path to config file.
}
\description{
Internal function to get the Lifebit Platform config file path.
Config file is stored in the user's R config directory as .cloudos_config.json
using tools::R_user_dir() for CRAN compliance.
}
\keyword{internal}
